import numpy as np
from open_ephys.analysis import Session
import glob
import os
import xml.etree.ElementTree as ETree
from ..utils import convert_port_number
import h5py

def load_neuropixel_configuration(data_dir, data_folder, ex_idx = 0, port_number = 1):
    '''
    get neuropixel probe information from xml condiguration files made by OpenEphys
    channel number and electrode x pos is sorded in the order of y pos when saved by openephys
    This function also sorts x pos and y pos in the order of channel number
    
    Args:
        data_dir (str): where to find the file
        data_folder (str): the xml file that describes neuropixel probe configuration
        ex_idx (int): experiment idx. This is usually 0.
        port_number (int): port number which a probe connected to. natural number from 1 to 4.
    
    Returns:
        config (dict) : dictionary thet contains electrode configuration
    '''

    data_path = os.path.join(data_dir, data_folder)
    xmlfile_path = glob.glob(os.path.join(data_path,'**/*.xml')) # take xml file path
    tree = ETree.parse(xmlfile_path[ex_idx])
    root = tree.getroot()

    config_list = []
    for a in root.iterfind('SIGNALCHAIN'):
        for b in a.iterfind('PROCESSOR'):
            for c in b.iterfind('EDITOR'):
                for d in c.iterfind('NP_PROBE'):
                    config = d.attrib
                    for e in d.iterfind('CHANNELS'):
                        ch_number = np.array([int(a[2:]) for a in e.attrib.keys()]) # only extract number from ch##
                        ch_bank = np.array([int(a) for a in e.attrib.values()])
                        config['channel'] = np.sort(ch_number)
                        config['ch_bank'] = ch_bank[np.argsort(ch_number)] # sort in the order of ch_number

                    for e in d.iterfind('ELECTRODE_XPOS'):
                        x_pos = np.array([int(a) for a in e.attrib.values()])
                        config['xpos'] = x_pos[np.argsort(ch_number)] # sort in the order of ch_number

                    for e in d.iterfind('ELECTRODE_YPOS'):
                        y_pos = np.array([int(a) for a in e.attrib.values()])
                        config['ypos'] = y_pos[np.argsort(ch_number)] # sort in the order of ch_number
                        
                    config_list.append(config)
                    
    return config_list[port_number-1] # Only take configuration that correspond to port nubmer

def load_neuropixel_data(data_dir, data_folder, datatype, node_idx=0, ex_idx=0, port_number=1):
    '''
    Load neuropixel data object and metadata. The data obeject has 4 properties of samples, sample_numbers, timestamps, and metadata.
    See this link: https://github.com/open-ephys/open-ephys-python-tools/tree/main/src/open_ephys/analysis
    
    Args:
        data_dir (str): data directory where the data files are located
        data_folder (str): data folder where 1 experiment data is saved
        datatype (str): datatype. 'ap' or 'lfp'
        node_idx (int): record node index. This is usually 0.
        ex_idx (int): experiment index. This is usually 0.
        port_number (int): port number which a probe connected to. natural number from 1 to 4. 
    
    Returns:
        tuple: Tuple containing:
            | **rawdata (object):** data object
            | **metadata (dict):** metadata
    '''
    
    if datatype == 'ap':
        datatype_idx = 2*(port_number-1) # even numbers correspond to action potential
    elif datatype == 'lfp':
        datatype_idx = 2*port_number-1 # odd numbers correspond to lfp
    else:
        raise ValueError(f"Unknown datatype {datatype}")
    
    # Load data and metadata
    data_path = os.path.join(data_dir, data_folder)
    session = Session(data_path)
    data = session.recordnodes[node_idx].recordings[ex_idx].continuous[datatype_idx]   
    metadata = data.metadata
    
    # Add electrode configuration to metadata
    config = load_neuropixel_configuration(data_dir, data_folder, ex_idx=ex_idx, port_number=port_number)
    metadata['slot'] = config['slot']
    metadata['port'] = config['port']
    metadata['dock'] = config['dock']
    metadata['channel'] = config['channel']
    metadata['ch_bank'] = config['ch_bank']
    metadata['xpos'] = config['xpos']
    metadata['ypos'] = config['ypos']
    metadata['referenceChannel'] = config['referenceChannel']
    metadata['probe_serial_number'] = config['probe_serial_number']
    metadata['probe_part_number'] = config['probe_part_number']
    metadata['bs_serial_number'] = config['bs_serial_number']
    metadata['bs_part_number'] = config['bs_part_number']
    metadata['bsc_serial_number'] = config['bsc_serial_number']
    metadata['bsc_part_number'] = config['bsc_part_number']
    metadata['headstage_serial_number'] = config['headstage_serial_number']
    metadata['headstage_part_number'] = config['headstage_part_number']
     
    if datatype == 'ap':
        metadata['apGainValue'] = config['apGainValue']
    else:
        metadata['lfpGainValue'] = config['lfpGainValue']
    
    return data, metadata
    
def load_neuropixel_event(data_dir, data_folder, datatype, node_idx=0, ex_idx=0, port_number=1):
    '''
    Load neuropixel's event data saved by openephys, accroding to datatype
    
    Args:
        data_dir (str): data directory where the data files are located
        data_folder (str): data folder where 1 experiment data is saved
        datatype (str): datatype. 'ap' or 'lfp'
        node_idx (int): record node index. This is usually 0.
        ex_idx (int): experiment index. This is usually 0.
        port_number (int): port number which a probe connected to. natural number from 1 to 4. 
    
    Returns:
        events (ndarray) : events data
    '''
    
    if (datatype != 'ap')&(datatype != 'lfp'):
        raise ValueError(f"Unknown datatype {datatype}")
    
    # Load data object
    data_path = os.path.join(data_dir, data_folder)
    session = Session(data_path)
    recording = session.recordnodes[node_idx].recordings[ex_idx]

    # Get probe name to search relevant events
    letter = chr(ord('@')+port_number) # convert port number to alphabetical value
    probe_name = f'Probe{letter}-{datatype.upper()}'
    
    # Initialize data structure for saving events
    ty = np.dtype([('line','i2'),('sample_number','i8'),('timestamp','f8'),('processor_id','i8'),
              ('stream_idx','i8'),('stream_name','S16'),('state','i8')])
    stream_name = recording.events.stream_name.values
    n_events = np.sum(stream_name==probe_name)
    events = np.zeros(n_events,dtype=ty)
    
    # Get event data relevant with probe name
    events['line'] = recording.events.line.values[stream_name==probe_name]
    events['sample_number'] = recording.events.sample_number.values[stream_name==probe_name]
    events['timestamp'] = recording.events.timestamp.values[stream_name==probe_name]
    events['processor_id'] = recording.events.processor_id.values[stream_name==probe_name]
    events['stream_idx'] = recording.events.stream_index.values[stream_name==probe_name]
    events['stream_name'] = recording.events.stream_name.values[stream_name==probe_name]
    events['state'] = recording.events.state.values[stream_name==probe_name]
    
    # Sort events according to sample_numbers (time_index)
    sort_idx = np.argsort(events['sample_number'])
    events['line'] = events['line'][sort_idx]
    events['sample_number'] = events['sample_number'][sort_idx]
    events['timestamp'] = events['timestamp'][sort_idx]
    events['processor_id'] = events['processor_id'][sort_idx]
    events['stream_idx'] = events['stream_idx'][sort_idx]
    events['stream_name'] = events['stream_name'][sort_idx]
    events['state'] = events['state'][sort_idx]
    
    return events

def get_neuropixel_digital_input_times(data_dir, data_folder, datatype, node_idx=0, ex_idx=0, port_number=1):
    '''
    Computes the times when sync line come to the degital channel in openephys.
    Openephys recodings doesn't always begin with 0 time index.
    
    Args:
        data_dir (str): data directory where the data files are located
        data_folder (str): data folder where 1 experiment data is saved
        datatype (str): datatype. 'ap' or 'lfp'
        node_idx (int): record node index. This is usually 0.
        ex_idx (int): experiment index. This is usually 0.
        port_number (int): port number which a probe connected to. Natural number from 1 to 4. 
    
    Returns:
        tuple: Tuple containing:
            | **on_times (n_times):** times at which sync line turned on
            | **off_times (n_times):** times at which sync line turned off
    '''
        
    # Load data and metadata
    data, metadata = load_neuropixel_data(data_dir, data_folder, datatype, node_idx=node_idx, ex_idx=ex_idx, port_number=port_number)
    FS = metadata['sample_rate']
    
    # Openephys recording doesn't always begin with 0 time. This initial_timestamp is necessary
    initial_timestamp = data.sample_numbers[0]/FS
    
    # Get on_times and off_times in digital data, based on initial_timestamps
    events = load_neuropixel_event(data_dir, data_folder, datatype, node_idx=0, ex_idx=0)
    on_times = events['sample_number'][events['state']==1]/FS - initial_timestamp
    off_times = events['sample_number'][events['state']==0]/FS - initial_timestamp
    
    return on_times, off_times

def load_ks_output(kilosort_dir, concat_data_dir, flag='spike'):
    '''
    load kilosort output preprocessed by kilosort.
    If falg is 'spike', it loads spike information (spike indices, spike label)
    If flag is 'template', it loads template informaiton that was used for spike sorting in kilosort
    If flag is 'channel', it loads channel mapping information in kilosort
    If flag is 'rez', it loads rez.mat file, which contains drift map information
    
    Args:
        kilosort_dir (str): kilosort directory (ex. '/data/preprocessed/kilosort')
        concat_data_dir (str): data directory that contains concatenated data and kilosort_output (ex. '2023-06-30_Neuropixel_ks_affi_bottom_port1')
        flag (str, optional): Which data you load. 'spike','template','channel', or 'rez'
        
    Return:
        kilosort_output (dict): preprocessed data by kilosort
    '''

    # Get kilosort directory path
    data_path = os.path.join(kilosort_dir, concat_data_dir)
    
    # path for kilosort data
    if flag == 'spike':
        spike_times_path = glob.glob(os.path.join(data_path,'**/spike_times.npy'),recursive=True)[0]
        spike_clusters_path = glob.glob(os.path.join(data_path,'**/spike_clusters.npy'),recursive=True)[0]
        ks_label_path = glob.glob(os.path.join(data_path,'**/cluster_KSLabel.tsv'),recursive=True)[0]
        amplitudes_path = glob.glob(os.path.join(data_path,'**/amplitudes.npy'),recursive=True)[0]
        cluster_amplitude_path = glob.glob(os.path.join(data_path,'**/cluster_Amplitude.tsv'),recursive=True)[0]
        cluster_group_path = glob.glob(os.path.join(data_path,'**/cluster_group.tsv'),recursive=True)[0]
    elif flag == 'template':
        spike_templates_path = glob.glob(os.path.join(data_path,'**/spike_templates.npy'),recursive=True)[0]
        template_features_path = glob.glob(os.path.join(data_path,'**/template_features.npy'),recursive=True)[0]
        template_feature_ind_path = glob.glob(os.path.join(data_path,'**/template_feature_ind.npy'),recursive=True)[0]
        templates_path = glob.glob(os.path.join(data_path,'**/templates.npy'),recursive=True)[0]
    elif flag == 'channel':
        channel_map_path = glob.glob(os.path.join(data_path,'**/channel_map.npy'),recursive=True)[0]
        channel_positions_path = glob.glob(os.path.join(data_path,'**/channel_positions.npy'),recursive=True)[0]
    elif flag == 'rez':
        rez_path = glob.glob(os.path.join(data_path,'**/rez.mat'),recursive=True)[0]
    
    # load kilsort data
    kilosort_output = {}
    if flag == 'spike':
        kilosort_output['spike_indices'] = np.load(spike_times_path)
        kilosort_output['spike_clusters'] = np.load(spike_clusters_path)
        kilosort_output['ks_label'] = np.genfromtxt(ks_label_path,skip_header=1,dtype=h5py.special_dtype(vlen=str))
        kilosort_output['amplitudes'] = np.load(amplitudes_path)
        kilosort_output['cluster_amplitude'] = np.genfromtxt(cluster_amplitude_path,skip_header=1,dtype=h5py.special_dtype(vlen=str))
        kilosort_output['cluster_group'] = np.genfromtxt(cluster_group_path,skip_header=1,dtype=h5py.special_dtype(vlen=str))
    elif flag == 'template':
        kilosort_output['spike_templates'] = np.load(spike_templates_path)
        kilosort_output['template_features'] = np.load(template_features_path)
        kilosort_output['template_feature_ind'] = np.load(template_feature_ind_path)
        kilosort_output['templates'] = np.load(templates_path)
    elif flag == 'channel':
        kilosort_output['channel_map'] = np.load(channel_map_path)
        kilosort_output['channel_positions'] = np.load(channel_positions_path)
    elif flag == 'rez':
        f = h5py.File(rez_path,'r')
        kilosort_output = dict(f['rez'])
    
    return kilosort_output

def load_parsed_ksdata(kilosort_dir, data_dir):
    '''
    load kilosort data (spike indices, clusters, and label) parsed into the task entries
    This data is not still synchronized
    
    Args:
        kilosort_dir (str): kilosort directory (ex. '/data/preprocessed/kilosort')
        data_dir (str): data directory that contains parsed data (ex. '2023-06-30_Neuropixel_ks_affi_bottom_port1_9847')
        
    Returns:
        spike_indices (nspikes): spike indices detected by kilosort (not spike times)
        spike_clusters (nspikes): unit label detected  by kilsort
    '''
    
    # Path for loading spikes and clusters
    data_path = os.path.join(kilosort_dir, data_dir)
    spike_path = os.path.join(data_path,'spike_indices_entry.npy')
    cluster_path = os.path.join(data_path,'spike_clusters_entry.npy')
    label_path = os.path.join(data_path,'ks_label.npy')

    # Load spikes and clusters
    spike_indices = np.load(spike_path)
    spike_clusters = np.load(cluster_path)
    ks_label = np.load(label_path)
    
    return spike_indices, spike_clusters, ks_label

def get_channel_bank_name(ch_bank_data, ch_config_dir ='/data/channel_config_np', filename='channel_bank.npy'):
    '''
    Get the information about which channels are used for recording. This function assumes channel configuration is either of below,
    long-br, middle, long-tr, top, long-tl, long-bl, bottom.
    
    Args:
        ch_bank_data (nch): channel bank information contained in neuropixel
        ch_config_dir (str, optional): directory that contains the channel configuration file
        filename (str, optional): filename that includes all bank information.
        
    Returns:    
        chname (str): channel name (long-br, middle, long-tr, top, long-tl, long-bl, bottom)
    '''
    data_path = os.path.join(ch_config_dir, filename)

    # Load data about channel configuration
    ch_bank_info =  np.load(data_path, allow_pickle=True).item()
    ch_bank_list = list(ch_bank_info.keys())

    # Compare actual data with configuration files to see whch bank of each channel is used.
    for chname in ch_bank_list:
        if np.all(ch_bank_info[chname] == ch_bank_data):
            break
        
    return chname



# kfdecoder.py
#
# Kalman Filter implementation

import numpy as np
from numpy.linalg import inv as inv # used in Kalman Filter

class KFDecoder(object):
    """
    Kalman Filter Decoder
    
    Args:
        C (float, optional): default 1
            This parameter scales the noise matrix associated with the transition in kinematic states.
            It effectively allows changing the weight of the new neural evidence in the current update.
            Our implementation of the Kalman filter for neural decoding is based on that of Wu et al 2003 (https://papers.nips.cc/paper/2178-neural-decoding-of-cursor-motion-using-a-kalman-filter.pdf)
            with the exception of the addition of the parameter C.
            The original implementation has previously been coded in Matlab by Dan Morris (http://dmorris.net/projects/neural_decoding.html#code)
    
    Attributes:
        model ([A, W, H, Q] list): list of matrices:
            | [State Transition model,
            | Covariance of State Transition model,
            | Observation model,
            | Covariance of Observation model]
    """

    def __init__(self, C=1):
        self.C = C

    def fit(self, X_kf_train, y_train):
        """
        Train Kalman Filter Decoder
        
        Args:
            X_kf_train (ntime , nfeatures): This is the neural data in Kalman filter format. See example file for an example of how to format the neural data correctly.
            y_train (ntime , noutputs): These are the outputs that are being predicted

        Calculations for :math:`A`, :math:`W`, :math:`H`, :math:`Q` are as follows:

        .. math:: 
            
            A = X2*X1' (X1*X1')^{-1}

        .. math:: 
            
            W = \\frac{(X_2 - A*X_1)(X_2 - A*X_1)'}{(timepoints - 1)}
        
        .. math:: 
        
            H = Y*X'(X*X')^{-1}
        
        .. math:: 
            
            Q = \\frac{(Y-HX)(Y-HX)' }{time points}
        """

        # Renaming and reformatting variables to be in a more standard kalman filter nomenclature (from Wu et al, 2003):
        # xs are the state (here, the variable we're predicting, i.e. y_train)
        # zs are the observed variable (neural data here, i.e. X_kf_train)
        X = np.matrix(y_train.T)
        Z = np.matrix(X_kf_train.T)

        # number of time bins
        nt = X.shape[1]

        # Calculate the transition matrix (from x_t to x_t+1) using least-squares, and compute its covariance
        # In our case, this is the transition from one kinematic state to the next
        X2 = X[:, 1:]
        X1 = X[:, 0:nt - 1]

        A = X2 * X1.T * inv(X1 * X1.T)  # Transition matrix
        W = (X2 - A * X1) * (X2 - A * X1).T / (
                    nt - 1) / self.C  # Covariance of transition matrix. Note we divide by nt-1 since only nt-1 points were used in the computation (that's the length of X1 and X2). We also introduce the extra parameter C here.

        # Calculate the measurement matrix (from x_t to z_t) using least-squares, and compute its covariance
        # In our case, this is the transformation from kinematics to spikes
        H = Z * X.T * (inv(X * X.T))  # Measurement matrix
        Q = ((Z - H * X) * ((Z - H * X).T)) / nt  # Covariance of measurement matrix

        params = [A, W, H, Q]
        print('Shape of State Transition model (A) :' + str(A.shape))
        print('Shape of Covariance of State Transition model :' + str(W.shape))
        print('Shape of Observation model (H) :' + str(H.shape))
        print('Shape of Covariance of Observation model :' + str(Q.shape))
        self.model = params

    def fit_awf(self, X_kf_train, y_train, A, W):
        """
        Train Kalman Filter Decoder with A and W fixed. A is the state transition model and W is the associated covariance

        
        Args:
            X_kf_train (ntime , nfeatures): This is the neural data in Kalman filter format. See example file for an example of how to format the neural data correctly
            y_train (ntime , noutputs): These are the outputs that are being predicted

        Calculations as follows:

        .. math::
            
            H = Y*X'(X*X')^{-1}

        .. math:: 
        
            Q = \\frac{(Y-HX)(Y-HX)' }{time points}
        """

        # Renaming and reformatting variables to be in a more standard kalman filter nomenclature (from Wu et al, 2003):
        # xs are the state (here, the variable we're predicting, i.e. y_train)
        # zs are the observed variable (neural data here, i.e. X_kf_train)
        X = np.matrix(y_train.T)
        Z = np.matrix(X_kf_train.T)

        # number of time bins
        nt = X.shape[1]

        # Calculate the transition matrix (from x_t to x_t+1) using least-squares, and compute its covariance
        # In our case, this is the transition from one kinematic state to the next
        X2 = X[:, 1:]
        X1 = X[:, 0:nt - 1]

        # A=X2*X1.T*inv(X1*X1.T) #Transition matrix
        # W=(X2-A*X1)*(X2-A*X1).T/(nt-1)/self.C #Covariance of transition matrix. Note we divide by nt-1 since only nt-1 points were used in the computation (that's the length of X1 and X2). We also introduce the extra parameter C here.

        # Calculate the measurement matrix (from x_t to z_t) using least-squares, and compute its covariance
        # In our case, this is the transformation from kinematics to spikes
        H = Z * X.T * (inv(X * X.T))  # Measurement matrix
        Q = ((Z - H * X) * ((Z - H * X).T)) / nt  # Covariance of measurement matrix

        print('Shape of State Transition model (A) :' + str(A.shape))
        print('Shape of Covariance of State Transition model :' + str(W.shape))
        print('Shape of Observation model (H) :' + str(H.shape))
        print('Shape of Covariance of Observation model :' + str(Q.shape))
        params = [A, W, H, Q]
        self.model = params

    def predict(self, X_kf_test, y_test):
        """
        Predict outcomes using trained Kalman Filter Decoder

        Args:
        X_kf_test (ntime, nfeatures): This is the neural data in Kalman filter format.
        y_test (ntime , noutputs): The actual outputs. This parameter is necesary for the Kalman filter (unlike other decoders) because the first value is nececessary for initialization
        
        Returns:
            y_test_predicted (ntime, noutputs): The predicted outputs
        """

        # Extract parameters
        A, W, H, Q = self.model

        # First we'll rename and reformat the variables to be in a more standard kalman filter nomenclature (I am following Wu et al):
        # xs are the state (here, the variable we're predicting, i.e. y_train)
        # zs are the observed variable (neural data here, i.e. X_kf_train)
        X = np.matrix(y_test.T)
        Z = np.matrix(X_kf_test.T)

        # Initializations
        num_states = X.shape[0]  # Dimensionality of the state
        states = np.empty(
            X.shape)  # Keep track of states over time (states is what will be returned as y_test_predicted)
        P_m = np.matrix(np.zeros([num_states, num_states]))  # This is a priori estimate of X covariance
        P = np.matrix(np.zeros([num_states, num_states]))  # This is a posteriori estimate of X covariance
        state = X[:, 0]  # Initial state
        states[:, 0] = np.copy(np.squeeze(state))

        # Get predicted state for every time bin
        for t in range(X.shape[1] - 1):
            # Do first part of state update - based on transition matrix
            P_m = A * P * A.T + W  # a priori estimate of x covariance ( P(k) = A*P(k-1)*A' + W )
            state_m = A * state  # a priori estimate of x ( X(k|k-1) = A*X(k-1) )

            # Do second part of state update - based on measurement matrix
            K = P_m * H.T * inv(H * P_m * H.T + Q)  # Calculate Kalman gain ( K = P_ap*H'* inv(H*P_ap*H' + Q) )
            P = (np.matrix(np.eye(num_states)) - K * H) * P_m  # (a posteriori estimate, P (I - K*H)*P_ap )
            state = state_m + K * (Z[:,t + 1] - H * state_m)  # compute a posteriori estimate of x (X(k) = X(k|k-1) + K*(Z - H*X(k|k-1))
            states[:, t + 1] = np.squeeze(state)  # Record state at the timestep
        y_test_predicted = states.T
        return y_test_predicted



# __init__.py
import socket
from . import analysis, data, postproc, precondition, preproc, visualization, tutorial_functions, utils

# Set the default memory limit
if socket.gethostname() == 'crab-eating':
    utils.set_memory_limit_gb(100)



# utils.py
# Any extra utility functions belong here
# Helper functions, math, other things that don't really pertain to neural data analysis

import numpy as np
import re
from datetime import datetime
import os
import sys
import math
from scipy.ndimage.interpolation import shift


'''
Test signals
'''
def generate_test_signal(duration, samplerate, frequencies, amplitudes, noise_amplitude=0.):
    '''
    Generates a test time series signal with multiple frequencies, specified in freq, for T timelength at a sampling rate of fs

    Args:
        duration (float): time period in seconds
        samplerate (int): sampling frequency in Hz
        frequencies (1D array): list of frequencies to be mixed in the test signal
        amplitudes (1D array): list of amplitudes for each frequency
        noise_amplitude (float, optional): amplitude of noise added on top of test signal

    Returns:
        tuple: Tuple containing:
            | **x (1D array):** cosine wave with multiple frequencies (and noise)
            | **t (1D array):** time vector for x
    '''
    n_samples = int(duration * samplerate)
    t = np.linspace(0, duration, n_samples, endpoint=False)

    x = np.random.normal(0,noise_amplitude,n_samples) # start with some noise

    for i in range(len(frequencies)):
        x += amplitudes[i] * np.cos(2 * np.pi * frequencies[i] * t)

    return x, t

def generate_multichannel_test_signal(duration, samplerate, n_channels, frequency, amplitude):
    '''
    Generate sine waves offset in phase by 2*pi/n_channels at the given amplitude and frequency

    Args: 
        duration (float): time in seconds
        samplerate (int): sampling rate of the signal in Hz
        n_channels (int): number of channels to generate
        frequency (float): frequency in Hz
        amplitude (float): amplitude of each sine wave

    Returns:
        (nt, nch) array: timeseries data across channels
    '''    
    time = np.arange(0, duration, 1/samplerate)
    data = []
    for i in range(n_channels):
        theta = 2*i*np.pi/n_channels # shift phase for each channel
        sinewave = amplitude * np.sin(2 * np.pi * frequency * time + theta)
        data.append(sinewave)
    data = np.array(data).T

    return data

def save_test_signal_ecube(data, save_dir, voltsperbit, datasource='Headstages'):
    '''
    Create a binary file with eCube formatting using the given data

    Args:
        data (nt, nch): test_signal to save
        save_dir (str): where to save the file
        voltsperbit (float): gain of the data you are creating
        datasource (str): eCube source from which you want the data to be 
            labeled (i.e. Headstages, AnalogPanel, or DigitalPanel)

    Returns:
        str: filename of the new data
    '''
    intdata = np.array(data/voltsperbit, dtype='<i2') # turn into integer data
    flatdata = intdata.reshape(-1)
    timestamp = np.array([1, 2, 3, 4], dtype='<i2')

    # Save it to the test file
    datestr = datetime.now().strftime("%Y-%m-%d_%H-%M-%S") # e.g. 2021-05-06_11-47-02
    filename = f"{datasource}_{data.shape[1]}_Channels_int16_{datestr}.bin"
    filepath = os.path.join(save_dir, filename)
    with open(filepath, 'wb') as f:
        f.write(timestamp)
        f.write(flatdata.tobytes())

    return filename

def count_unique_symbols(files):
    '''
    Utility for counting how many times each unique symbol is listed in the given list
    and ranking them by descending number of uses.

    Args:
        files (list): list of filenames containing symbols generated by vscode 'List Symbols' 

    Returns:
        tuple: tuple containing:
            | **unique_symbols (list):** list of unique symbols
            | **counts (list):** list of counts for each unique symbol
    '''
    symbols = []
    for filename in files:
        with open(filename, mode='r') as f:
            text = f.read()
        symbols += re.findall("variable (\w+)", text)
    unique_symbols, counts = np.unique(symbols, return_counts=True)
    order = np.argsort(-counts)

    return unique_symbols[order], counts[order]

'''
Digital calc
'''
def convert_analog_to_digital(analog_data, thresh=.3):
    '''
    This function takes analog data and converts it to digital data given a 
    threshold. It scales the analog to between 0 and 1 and uses thres as a 

    Args: 
        analog_data (nt, nch): Time series array of analog data
        thresh (float, optional): Minimum threshold value to use in conversion

    Returns:
        (nt, nch): Array of 1's or 0's indicating if the analog input was above threshold  
    '''
    # Scale data between 0 and 1 so that threshold is a percentange
    minval = np.min(analog_data)
    maxval = np.max(analog_data)

    analog_data_scaled = (analog_data - minval)/(maxval - minval)

    # Initialize digital_data
    digital_data = np.zeros(analog_data_scaled.shape, dtype='bool') # Default to zero

    # Set any value greater than threshold to be 1
    digital_data[analog_data_scaled >= thresh] = 1

    return digital_data

def detect_edges(digital_data, samplerate, rising=True, falling=True, check_alternating=True, 
        min_pulse_width=None):
    '''
    Finds the timestamp and corresponding value of all the bit flips in data. Assumes 
    the first element in data isn't a transition

    By default, also enforces that rising and falling edges must alternate, always taking the
    last edge as the most valid one. For example::

        >>> data = [0, 0, 3, 0, 3, 2, 2, 0, 1, 7, 3, 2, 2, 0]
        >>> ts, values = detect_edges(data, fs)
        >>> print(values)
        [3, 0, 3, 0, 7, 0]

    Args:
        digital_data (ntime x 1): masked binary data array
        samplerate (int): sampling rate of the data used to calculate timestamps
        rising (bool, optional): include low to high transitions
        falling (bool, optional): include high to low transitions
        check_alternating (bool, optional): if True, enforces that rising and falling
            edges must be alternating. An edge is valid when it is no longer rising or falling.
        min_pulse_width (float, optional): if not None, makes sure rising edges are
            followed by a minimum pulse width before calculating edge values

    Returns:
        tuple: tuple containing:
            | **timestamps (nbitflips):** when the bits flipped
            | **values (nbitflips):** corresponding values for each change
    '''
    digital_data = np.squeeze(np.uint64(digital_data)) # important conversion for binary math
    if min_pulse_width:
        channels = convert_digital_to_channels(digital_data)
        channels = copy_edges_forwards(channels, int(samplerate*min_pulse_width)-1, axis=0)
        digital_data = convert_channels_to_digital(channels)

    # Find edges
    rising_idx = (~digital_data[:-1] & digital_data[1:]) > 0 # find low->high transitions
    falling_idx = (~digital_data[1:] & digital_data[:-1]) > 0

    # Find any non-alternating edges
    invalid = np.zeros((len(digital_data)-1,), dtype='?')
    if check_alternating:
        all_edges = np.where(rising_idx | falling_idx)[0]
        next_edge_rising = True
        next_edge_falling = True
        for idx in range(len(all_edges)):
            this_idx = all_edges[idx]
            if next_edge_rising and rising_idx[this_idx]:
                # Expected rising and found rising
                next_edge_rising = False 
                next_edge_falling = True
            elif next_edge_falling and falling_idx[this_idx]:
                # Expected falling and found falling
                next_edge_rising = True 
                next_edge_falling = False
            elif idx > 0: # skip the first edge since there is no previous edge
                # Unexpected; there must be an extra edge somewhere.
                # We will count this one as valid and the previous one as invalid
                prev_idx = all_edges[idx-1]
                invalid[prev_idx] = True
    
    # Assemble final index    
    logical_idx = np.zeros((len(digital_data)-1,), dtype='?')
    if rising:
        logical_idx |= rising_idx
    if falling:
        logical_idx |= falling_idx
    logical_idx &= np.logical_not(invalid)
    logical_idx = np.insert(logical_idx, 0, False) # first element never a transition

    time = np.arange(np.size(digital_data))/samplerate
    return time[logical_idx], digital_data[logical_idx]

def extract_bits(data, mask):
    '''
    Apply bit mask and shift data to the least significant set bit in the mask. 
    For example,
    extract_bits(0001000011110000, 1111111100000000) => 00010000
    extract_bits(0001000011110000, 0000000011111111) => 11110000
    extract_bits(0001000011001100, 0000001111001111) => 00111100

    Args:
        data (ntime): digital data
        mask (int): which bits to filter

    Returns:
        (nt): masked and shifted data
    '''
    data = np.array(data).astype(int)
    result = np.zeros(shape=data.shape, dtype=data.dtype)
    mask_bit_positions = [i for i, bit in enumerate(reversed(bin(mask)[2:])) if bit == '1']

    for pos in mask_bit_positions[::-1]:
        result <<= 1
        result |= ((data >> pos) & 1).astype(data.dtype)

    return result

def convert_channels_to_mask(channels):
    '''
    Helper function to take a range of channels into a bitmask

    Args:
        channels (int array): 0-indexed channels to be masked
    
    Returns:
        int: binary mask of the given channels
    '''
    try:
        # Range of channels
        _ = iter(channels)
        channels = np.array(channels)
        flags = np.zeros(64, dtype=int)
        flags[channels] = 1
        return int(np.dot(np.array([2**i for i in range(64)]), flags))
    except:
        
        # Single channel
        return int(1 << channels)

def convert_digital_to_channels(data_64_bit):
    '''
    Converts 64-bit digital data from eCube into channels.

    Args:
        data_64_bit (n): masked 64-bit data, little-endian

    Returns:
        (n, 64): where channel 0 is least significant bit
    '''

    # Take the input, split into bytes, then unpack each byte, all little endian
    packed = np.squeeze(np.uint64(data_64_bit)) # required conversion to unsigned int
    unpacked = np.unpackbits(packed.view(np.dtype('<u1')), bitorder='little')
    return unpacked.reshape((packed.size, 64))

def convert_channels_to_digital(data_channels):
    '''
    Converts binary channels from eCube into 64-bit digital data.

    Args:
        data_channels (n, 64): where channel 0 is least significant bit

    Returns:
        (n): masked 64-bit data, little-endian
    '''

    # Take the input, split into bytes, then unpack each byte, all little endian
    packed = np.packbits(data_channels, bitorder='little').view(np.dtype('<u8'))
    return np.squeeze(packed)
    
def get_edges_from_onsets(onsets, pulse_width):
    '''
    This function calculates the values and timepoints corresponding to a given time series 
    of pulse onsets (timestamp corresponding to the rising edge of a pulse). 
    Args:
        onsets (nonsets): Time point corresponding to a pulse onset. 
        pulse_width (float): Pulse duration 
    Returns:
        tuple: tuple containing:
            | **timestampes (2*nonsets + 1):** Timestamps of the rising and falling edges. Always starts at 0.
            | **values (2*nonsets + 1):** Values corresponding to the output timestamps.
    '''
    timestamps = np.zeros((1+len(onsets)*2,))
    values = np.zeros((1+len(onsets)*2,))
    for t in range(len(onsets)):
        timestamps[1+2*t] = onsets[t]
        values[1+2*t] = 1
        timestamps[2+2*t] = onsets[t]+pulse_width
        values[2+2*t] = 0
    return timestamps, values

def get_pulse_edge_times( digital_data, samplerate ):

    """get_pulse_edge_times

    Args:
        digital_data (nt, 1): array of data from ecube digital panel
        samplerate (numeric): data sampling rate (Hz)

    Returns:
        edge_times (npulse, 2): start and end times from each detected pulse
    """

    edge_times, edge_val = detect_edges(digital_data, samplerate)
    start_idx = np.where(edge_val == 1)[0][0]
    end_idx = np.where(edge_val == 0)[0][-1]
    edge_times = edge_times[start_idx:(end_idx+1)]
    edge_val = edge_val[start_idx:(end_idx+1)]
    edge_pairs = edge_times.reshape(-1,2)

    return edge_pairs

def compute_pulse_duty_cycles( edge_pairs ):

    """compute_pulse_duty_cycles

    Args:
        edge_pairs (npulse, 2): start, end times from a series of pulses

    Returns:
        duty_cycle (npulse): duty cycle of each pulse. Pulse period assumed to be constant.
    """

    pulse_times = edge_pairs[:,0]
    pulse_period = np.diff(pulse_times,axis=0)
    duty_cycle = np.squeeze(np.diff(edge_pairs,axis=-1))/pulse_period[0]

    return duty_cycle

def max_repeated_nans(a):
    '''
    Utility to calculate the maximum number of consecutive nans

    Args:
        a (ndarray): input sequence

    Returns:
        int: max consecutive nans
    '''
    mask = np.concatenate(([False],np.isnan(a),[False]))
    if ~mask.any():
        return 0
    else:
        idx = np.nonzero(mask[1:] != mask[:-1])[0]
        return (idx[1::2] - idx[::2]).max()

def copy_edges_forwards(data, n_steps, truncate_edges=False, copy_per_step=False, axis=0):
    '''
    Forces pulses to have a fixed width of eactly n_steps. First, find the rising edges of the data,
    then copy them forwards n_steps times. Works across multiple channels simulatenously.

    Args:
        data ((nt,) or (nt, nch)): digital data
        n_steps (int): how many timesteps should pulses be
        truncate_edges (bool, optional): if True, then edges will always be set to n_steps length. 
            If false, then edges that are longer than n_steps will remain the same length. Default False.
        copy_per_step (bool, optional): copy edges one step at a time or one edge at a time; changes
            processing time but output stays the same. If there are long edges, the default option False
            is faster. If there are a lot of short edges, then setting copy_per_step=True will be faster. 
            Default False.
        axis (int, optional): along which axis to copy edges. Default 0.

    Returns:
        (nt,) or (nt, nch): digital data but with fixed pulse widths

    Note: 
        Only works for 1- or 2-D arrays
    '''
    if data.ndim == 1:
        data = np.expand_dims(data, 1)
    if axis != 0:
        data = data.T
    nt = data.shape[0]
    edges = (~data[:-1] & data[1:]) > 0 # find low->high transitions
    edges = np.insert(edges, 0, False, axis=0) # first element never a transition
    if not copy_per_step:
        # Find the edges, set their pulses directly
        idx = np.where(edges)
        for e_idx in zip(idx[0],idx[1]):
            edges[e_idx[0]:min(nt,e_idx[0]+n_steps+1),e_idx[1]] = 1
    else:
        # For all of the data at once, shift to the right by n_steps, or-ing with edges at each step
        for n in range(n_steps):
            edges_shifted = np.roll(edges, 1, axis=0)
            edges_shifted[0,:] = 0
            edges = np.logical_or(edges, edges_shifted)
    if axis != 0:
        edges = edges.T
        data = data.T
    if truncate_edges:
        return np.squeeze(edges)
    return np.squeeze(edges | data)

def count_repetitions(arr, diff_thr=0):
    """
    Counts the number of repetitions in an array. Always counts the first and  
    last element of the array as different from before and after the array.

    Args:
        arr (nt,): The input array. Only supports 1d arrays.
        diff_thr (numeric, optional): Minimum step size in the data.

    Returns:
        tuple: A tuple of two numpy arrays:
            | **repetitions (nt,):** Lengths of the repetitions in the input array, 
            | **change_idx (nt,):** Indices where the repetitions start
    """
    assert np.array(arr).ndim == 1, "Only supports 1d arrays"
    if len(arr) == 0:
        return [], []

    # Calculate the absolute difference between adjacent elements in the array
    diff = np.abs(np.diff(arr))
    change = diff > diff_thr
    change = np.insert(change, 0, True)

    # Find the indices where the 'change' array is True
    change_idx = np.where(change)[0]

    # Count the length of the gaps between those changes to find the repetitions
    change = np.append(change_idx, len(arr))
    repetitions = np.diff(change)

    return repetitions, change_idx


def segment_array(arr, category, duplicate_endpoints=False):
    '''
    Segments an array into subarrays based on a corresponding category array.

    Args:
        arr (nt,): The array to segment.
        category (nt,): An array of the same length as `arr` containing a 
            category label for each element in the corresponding array.
        duplicate_endpoints (bool, optional): if True, each subsequent subarray
            will start with the last element of the preceding subarray.

    Returns:
        tuple: Tuple containing:
            | **segments (list of arrays):** A list of subarrays of `arr`, where each subarray corresponds to a unique value in `category`.
            | **segmented_category (list of arrays):** An array of the same length as `segments`, where each element corresponds to the category label for the corresponding subarray in `segments`.
    '''
    run_arr, idx = count_repetitions(category)
    end_indices = np.cumsum(run_arr)
    if duplicate_endpoints:
        segments = []
        start_idx = 0
        for end_idx in end_indices:
            if start_idx == 0:
                segments.append(arr[:end_idx])
            else:
                segments.append(arr[start_idx-1:end_idx])
            start_idx = end_idx
    else:
        segments = np.split(arr, end_indices[:-1])
    return segments, [category[i] for i in idx]

def count_repetitions(arr, diff_thr=0):
    """
    Counts the number of repetitions in an array. Always counts the first and  
    last element of the array as different from before and after the array.

    Args:
        arr (nt,): The input array. Only supports 1d arrays.
        diff_thr (numeric, optional): Minimum step size in the data.

    Returns:
        tuple: A tuple of two numpy arrays:
        | **repetitions (nt,):** Lengths of the repetitions in the input array, 
        | **change_idx (nt,):** Indices where the repetitions start
    """
    assert np.array(arr).ndim == 1, "Only supports 1d arrays"
    if len(arr) == 0:
        return [], []

    # Calculate the absolute difference between adjacent elements in the array
    diff = np.abs(np.diff(arr))
    change = diff > diff_thr
    change = np.insert(change, 0, True)

    # Find the indices where the 'change' array is True
    change_idx = np.where(change)[0]

    # Count the length of the gaps between those changes to find the repetitions
    change = np.append(change_idx, len(arr))
    repetitions = np.diff(change)

    return repetitions, change_idx

def nextpow2(x):
    '''
    Next higher power of 2. It is often useful for finding the nearest power of two 
    sequence length for FFT operations.

    Args:
        x (int or float): input number 
        
    Returns:
        int: the first P such that 2**P >= abs(x).
    '''
    return 1 if x == 0 else math.ceil(math.log2(x))

'''
Other utils
'''
# simple progressbar, not tied to the iterator
def print_progress_bar(count, total, status=''):
    """print_progress_bar

    Args:
        count (num): current progress count
        total (int): total count, i.e. what count is at 100%
        status (str, optional): printed status message. Defaults to ''.
    """
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()

def derivative(x, y, norm=True):
    '''
    Computes the derivative of y along x.

    Args:
        x (nt): independent variable, e.g. time
        y (nt, ...): dependent variable, e.g. position
        norm (bool, optional): also compute the norm of y if it is multidimensional (default True). Set to false to output component wise derivative. 

    Returns:
        nt: derivative of y
    '''
    dy = np.gradient(y, axis=0, edge_order=2)
    dx = np.gradient(x)
    if norm and dy.ndim > 1:
        dy = np.linalg.norm(dy, axis=1)
        dydx = dy/dx
    elif not norm and dy.ndim > 1:
        dydx = dy/np.expand_dims(dx, len(dy.shape)-1)
    else:
        dydx = dy/dx

    return dydx

def calc_euclid_dist_mat(pos):
    '''
    Calculates a matrix of euclidean distance. Each entry in the matrix is the distance between ith and jth position
    
    Args:
        pos (nch,2): x, y position list, e.g. for each electrode
    
    Returns:
        (nch, nch) array: distance between each given position
    
    '''
    n_channels = np.shape(pos)[0]
    dist = np.zeros((n_channels, n_channels))
    x_pos = pos[:,0]
    y_pos = pos[:,1]
    for index, _ in np.ndenumerate(dist):
        p1 = index[0]
        p2 = index[1]
        dist[p1, p2] = math.dist([x_pos[p1], y_pos[p1]], [x_pos[p2], y_pos[p2]])
    
    return dist
    
def calc_radial_dist(pos, origin=(0,0)):
    '''
    Calculates a matrix of radial distance from a given origin. Each entry in the matrix is the distance between ith and jth electrode channel
        
    Args:
        pos (nch,2): x, y position list, e.g. for each electrode
        origin (2,): point from which to calculate radial distance
    
    Returns:
        (nch,) array: radius between each given position and the origin
    '''
    n_channels = np.shape(pos)[0]
    dist = np.zeros((n_channels,))
    for i, p in enumerate(pos):
        dist[i] = math.dist(origin, p)
    
    return dist

def first_nonzero(arr, axis=0, all_zeros_val=-1):
    '''
    Helper function to find the first non-zero element in an array

    Args:
        arr (ndarray): array containing zeros
        axis (int, optional): axis along which to compute the first nonzero. Defaults to 0.
        all_zeros_val (float, optional): value to indicate no nonzero elements were found. Defaults to -1.

    Returns:
        ndarray: array of indices with one less dimension than the input
    '''
    mask = (arr != 0)
    return np.where(mask.any(axis=axis), mask.argmax(axis=axis), all_zeros_val)

'''
Synchronization
'''
def extract_barcodes_from_times(on_times,off_times,inter_barcode_interval=30,bar_duration=0.017,barcode_duration_ceiling=2,nbits=32,):
    """
    Read barcodes from timestamped rising and falling edges. This function came from the openephys repository
    
    Notes:
        ignores first code in prod (ok, but not intended)
        ignores first on pulse (intended - this is needed to identify that a barcode is starting)
    
    Args:
        on_times (ndarray): Timestamps of rising edges on the barcode line
        off_times (ndarray): Timestamps of falling edges on the barcode line
        inter_barcode_interval (float) : Minimun duration of time between barcodes.
        bar_duration (float): A value slightly shorter than the expected duration of each bar
        barcode_duration_ceiling (float) : The maximum duration of a single barcode
        nbits (int): The bit-depth of each barcode
    
    Returns:
        tuple: tuple containing:
            | **barcode_start_times (list):** For each detected barcode, the time at which that barcode started
            | **barcodes (list of int):** For each detected barcode, the value of that barcode as an integer.
    """

    start_indices = np.diff(on_times)
    a = np.where(start_indices > inter_barcode_interval)[0]
    barcode_start_times = on_times[a + 1]
    
    barcodes = []
    for i, t in enumerate(barcode_start_times):
        oncode = on_times[np.where(np.logical_and(on_times > t, on_times < t + barcode_duration_ceiling))[0]]
        offcode = off_times[np.where(np.logical_and(off_times > t, off_times < t + barcode_duration_ceiling))[0]]
        currTime = offcode[0]
        bits = np.zeros((nbits,))

        for bit in range(0, nbits):
            nextOn = np.where(oncode > currTime)[0]
            nextOff = np.where(offcode > currTime)[0]

            if nextOn.size > 0:
                nextOn = oncode[nextOn[0]]
            else:
                nextOn = t + inter_barcode_interval
            if nextOff.size > 0:
                nextOff = offcode[nextOff[0]]
            else:
                nextOff = t + inter_barcode_interval
            if nextOn < nextOff:
                bits[bit] = 1
            currTime += bar_duration

        barcode = 0

        # least sig left
        for bit in range(0, nbits):
            barcode += bits[bit] * pow(2, bit)

        barcodes.append(int(barcode))

    return barcode_start_times, barcodes

def get_first_last_times(barcode_on_times, barcode_on_times_main, barcode, barcode_main):
    '''
    Get the first and last time when barcodes (sync pulses) come to each stream.
    
    Args: 
        barcode_on_times (n_times) : the times at which barcode comes to the auxiliary stream
        barcode_on_times_main (k_times) : the times at which barcode comes to the main stream
        barcode (n-length list) : Unique barcode number in the auxiliary stream
        barcode_main (k-length list) : Unique barcode number in the main stream
    
    Return:
        tuple: tuple containing:
            | **first_last_times (2):** barcode on_times that corresponds to the first and last barcode in the recording
            | **first_last_times (2):** barcode on_times in the main stream that corresponds to the first and last barcode in the recording
    '''
    
    # Get barcode index that is shared across both streams
    index_main = []
    index = []
    
    if barcode_on_times.shape[0] == barcode_on_times_main.shape[0]:
        # if shape of barcode timing is same, use the first and last indices
        index_main.append(0)
        index_main.append(-1)
        index.append(0)
        index.append(-1)
    else:
        for idx, i_barcode in enumerate(barcode):
            if i_barcode in barcode_main:
                index_main.append(barcode_main.index(i_barcode))
                index.append(idx)
    
    # Get the indices that correspond to the first and last barcode in the recording
    first_last_idx = [index[0],index[-1]]
    first_last_idx_main = [index_main[0],index_main[-1]]
    
    # Get the times that correspond to the first and last barcode in the recording
    first_last_times = barcode_on_times[first_last_idx]
    first_last_times_main = barcode_on_times_main[first_last_idx_main]
    
    return first_last_times,first_last_times_main

def sync_timestamp_offline(timestamp, on_times, on_times_main):
    '''
    Synchroniza timestamps with timestamps in another stream
    
    Args
        timestamps (nt) : timestamps in the auxiliary stream that should be synchronized to main stream
        on_time (2) : the first and last times when sync pulses come to the auxiliary stream in the recording
        on_time_main (2) : the first and last times when sync pulses come to the main stream in the recording
    
    Retuen:
        tuple: tuple containing:
            | **sync_timestamps (nt):** synchronized timestamps
            | **scaling (float):** scaling factor between streams
    '''
    
    scaling = (on_times_main[-1] - on_times_main[0])/(on_times[-1] - on_times[0])
    sync_timestamp = (timestamp - on_times[0]) * scaling + on_times_main[0]
    
    return sync_timestamp, scaling

def convert_port_number(port_number, datatype='ap'):
    '''
    convert port_number to directory name made by openephys
    
    Args:
        port_number (int): port number which a probe connected to. natural number from 1 to 4.
        datatyoe (str, optional): datatype of neuropixel. 'ap' or 'lfp'
    
    Returns:
        probe_dir (str): Probe directory name that contains AP data
    '''
    letter = chr(ord('@')+port_number)
    probe_dir = f'Neuropix-PXI-100.Probe{letter}-{datatype.upper()}'
    
    return probe_dir


import csv
import os
from pandas import read_csv

def load_optitrack_metadata(data_dir, filename, metadata_row=0):
    '''
    This function loads optitrack metadata from .csv file that has 1 rigid body
    exported with the following settings:

        | **Markers:** Off
        | **Unlabeled markers:** Off
        | **Quality Statistics:** Off
        | **Rigid Bodies:** On
        | **Rigid Body Markers:** Off
        | **Bones:** Off
        | **Bone Markers:** Off
        | **Header Information:** On
        | **Optitrack format Version(s):** 1.23

    Required packages: csv, pandas

    Args:
        data_dir (string): Directory to load data from
        filename (string): File name to load within the data directory

    Returns:
        dict: Dictionary of metadata for for an optitrack datafile
    '''

    # Constants for indexing into mocap data
    mocap_data_column_idx = 2 # Column index where data begins
    rigid_body_name_idx_csvrow = 3 #csv row that contains the rigid body name
    column_type_idx_csvrow = 5 #csv row that contains the column type 
    column_names_idx_csvrow = 6  #csv row that contains the column names

    # Initialize empty dict to add to
    mocap_metadata = {}

    filepath = os.path.join(data_dir, filename)
    # Create dict with .csv meta data
    with open(filepath) as csvfile:
        mocap_reader = csv.reader(csvfile, delimiter= ',')  # csv.reader object for optitrack data loading

        for idx_csvrow, row in enumerate(mocap_reader):
            # Create a dictionary of metadata for the first line
            if idx_csvrow == metadata_row:
                # For every pair of values in the list add to dict
                for idx_dictval in range(0, len(row), 2):
                    mocap_metadata[row[idx_dictval]]= row[idx_dictval + 1]

            # Get Rigid body name for each column
            elif idx_csvrow == rigid_body_name_idx_csvrow:
                mocap_metadata['Rigid Body Name'] = row[mocap_data_column_idx:][0]

            # Load whether column is rotation/position
            elif idx_csvrow == column_type_idx_csvrow:
                mocap_metadata['Data Column Motion Types'] = row[mocap_data_column_idx:]

            # Get column names for data
            elif idx_csvrow == column_names_idx_csvrow:
                mocap_metadata['Data Column Motion Names'] = row[mocap_data_column_idx:]  

    if 'Export Frame Rate' in mocap_metadata:
        mocap_metadata['samplerate'] = float(mocap_metadata['Export Frame Rate'])
    return mocap_metadata

def load_optitrack_data(data_dir, filename):
    '''
    This function loads a series of x, y, z positional data from the optitrack
    .csv file that has 1 rigid body exported with the following settings:

        | **Markers:** Off
        | **Unlabeled markers:** Off
        | **Quality Statistics:** Off
        | **Rigid Bodies:** On
        | **Rigid Body Markers:** Off
        | **Bones:** Off
        | **Bone Markers:** Off
        | **Header Information:** On
        | **Optitrack format Version(s):** 1.23

    Required packages: pandas, numpy

    Args:
        data_dir (string): Directory to load data from
        filename (string): File name to load within the data directory

    Returns:
        tuple: Tuple containing:
            | **mocap_data_pos (nt, 3):** Positional mocap data
            | **mocap_data_rot (nt, 4):** Rotational mocap data
    '''

    # Load the metadata to check the columns are going to line up
    mocap_metadata = load_optitrack_metadata(data_dir, filename)
    if not mocap_metadata:
        raise Exception('No metadata found for optitrack file')
    assert mocap_metadata['Rotation Type'] == 'Quaternion', 'Rotation type must be Quaternion'
    assert mocap_metadata['Format Version'] == '1.23', 'Only supports version 1.23'

    # Load the data columns
    column_names_idx_csvrow = 5 # Header row index
    mocap_data_rot_column_idx = range(2,6) # Column index for rotation data
    mocap_data_pos_column_idx = range(6,9) # Column indices for position data
    filepath = os.path.join(data_dir, filename)
    # Load .csv file as a pandas data frame, convert to a numpy array, and remove
    # the 'Frame' and 'Time (Seconds)' columns.
    mocap_data_rot = read_csv(filepath, header=column_names_idx_csvrow).to_numpy()[:,mocap_data_rot_column_idx]
    mocap_data_pos = read_csv(filepath, header=column_names_idx_csvrow).to_numpy()[:,mocap_data_pos_column_idx]

    return mocap_data_pos, mocap_data_rot

def load_optitrack_time(data_dir, filename):
    '''
    This function loads timestamps from the optitrack .csv file 

    Required packages: pandas, numpy

    Args:
        data_dir (string): Directory to load data from
        filename (string): File name to load within the data directory

    Returns:
        (nt): Array of timestamps for each captured frame
    '''

    column_names_idx_csvrow = 5 # Header row index
    timestamp_column_idx = 1 # Column index for time data
    filepath = os.path.join(data_dir, filename)
    # Load .csv file as a pandas data frame, convert to a numpy array, and only
    # return the 'Time (Seconds)' column
    timestamps = read_csv(filepath, header=column_names_idx_csvrow).to_numpy()[:,timestamp_column_idx]
    return timestamps



'''
Interface between database methods/models and data analysis code
'''
import json
import os
import pickle
import sys
from collections import defaultdict
import warnings
import traceback
import pandas as pd
import numpy as np
try:
    from db import dbfunctions as bmi3d
    from db.tracker import dbq
except:
    warnings.warn("Database not configured")
    traceback.print_exc()

'''
Lookup
'''

# Some global variable shenanigans.
this = sys.modules[__name__]
this.DB_TYPE = 'bmi3d' # can be modified with `db.DB_TYPE` after `from aopy.data import db``
this.BMI3D_DBNAME = 'default' # can be modified with `db.BMI3D_DBNAME` after `from aopy.data import db`

def lookup_sessions(id=None, subject=None, date=None, task_name=None, task_desc=None, session=None, project=None, 
                    experimenter=None, exclude_ids=[], filter_fn=lambda x:True, **kwargs):
    '''
    Returns list of entries for all sessions on the given date

    Args:
        id (int or list, optional): Lookup sessions with the given ids, if provided.
        subject (str, optional): Lookup sessions with the given subject, if provided.
        date (multiple, optional): Lookup sessions from the given date, if provided. Accepts multiple formats:
            | datetime.date object
            | (start, end) tuple of datetime.date objects
            | (start, end) tuple of strings in the format 'YYYY-MM-DD'
            | (year, month, day) tuple of integers
        task_name (str, optional): Lookup sessions with the given task name, if provided. Examples include
            `manual control`, `tracking`, `nothing`, etc.
        task_desc (str, optional): Lookup sessions with the given task description, if provided. Examples include
            `flash`, `simple center out`, `resting state`, etc.
        session (str, optional): Lookup sessions with the given session name, if provided.
        project (str, optional): Lookup sessions with the given project name, if provided.
        experimenter (str, optional): Lookup sessions with the given experimenter, if provided.
        exclude_ids (list, optional): Exclude sessions with matching task entry ids, if provided.
        filter_fn (function, optional): Additional filtering, of signature `fn(session)->bool. 
            Defaults to `lambda x:True`.
        kwargs (dict, optional): optional keyword arguments to pass to database lookup function.

    Returns:
        list: list of TaskEntry sessions matching the query
    '''
    if this.DB_TYPE == 'bmi3d':
        if id is not None and isinstance(id, list):
            kwargs['pk__in'] = id
        elif id is not None:
            kwargs['pk'] = id
        if subject is not None:
            kwargs['subj'] = subject
        if date is not None:
            kwargs['date'] = date
        if task_name is not None:
            kwargs['task'] = task_name
        if task_desc is not None:
            kwargs['entry_name'] = task_desc
        if session is not None:
            kwargs['session'] = session
        if project is not None:
            kwargs['project'] = project
        if experimenter is not None:
            kwargs['experimenter__name'] = experimenter
        
        # Initial lookup, wrapped into BMI3DTaskEntry objects
        dbname = kwargs.pop('dbname', this.BMI3D_DBNAME)
        entries = bmi3d.get_task_entries(dbname=dbname, **kwargs)
        entries = [BMI3DTaskEntry(e) for e in entries]

        # Additional filtering
        if len(entries) == 0:
            warnings.warn("No entries found")
            return []
        return [e for e in entries if filter_fn(e) and e.id not in exclude_ids]
    else:
        warnings.warn("Unsupported db type!")
        return []
    
def lookup_flash_sessions(mc_task_name='manual control', **kwargs):
    '''
    Returns list of entries for all flash sessions on the given date.
    See :func:`~aopy.data.db.lookup_sessions` for details.
    '''
    filter_fn = kwargs.pop('filter_fn', lambda x:True) and (lambda te: 'flash' in te.task_desc)
    return lookup_sessions(task_name=mc_task_name, filter_fn=filter_fn, **kwargs)

def lookup_mc_sessions(mc_task_name='manual control', **kwargs):
    '''
    Returns list of entries for all manual control sessions on the given date
    See :func:`~aopy.data.db.lookup_sessions` for details.
    '''
    filter_fn = kwargs.pop('filter_fn', lambda x:True) and (lambda te: 'flash' not in te.task_desc)
    return lookup_sessions(task_name=mc_task_name, filter_fn=filter_fn, **kwargs)

def lookup_tracking_sessions(tracking_task_name='tracking', **kwargs):
    '''
    Returns list of entries for all tracking sessions on the given date
    See :func:`~aopy.data.db.lookup_sessions` for details.
    '''
    return lookup_sessions(task_name=tracking_task_name, **kwargs)

def lookup_bmi_sessions(bmi_task_name='bmi control', **kwargs):
    '''
    Returns list of entries for all bmi control sessions on the given date
    See :func:`~aopy.data.db.lookup_sessions` for details.
    '''
    return lookup_sessions(task_name=bmi_task_name, **kwargs)

def lookup_decoder_parent(task_name='nothing', task_desc='decoder parent', **kwargs):
    '''
    Lookup by project and session
    '''
    return lookup_sessions(task_name=task_name, task_desc=task_desc, **kwargs)

def lookup_decoders(id=None, parent_id=None, **kwargs):
    '''
    Returns list of decoders with the given filter parameters

    Args:
        id (int or list, optional): Lookup decoders with the given ids, if provided.
        parent_id (int, optional): Lookup decoders with the given parent ids, if provided.
        kwargs (dict, optional): optional keyword arguments to pass to database lookup function.

    Returns:
        list: list of Decoder records matching the query
    '''
    if id is not None and isinstance(id, list):
        kwargs['id__in'] = id
    elif id is not None:
        kwargs['id'] = id
    if parent_id is not None:
        kwargs['entry_id'] = parent_id

    dbname = kwargs.pop('dbname', this.BMI3D_DBNAME)
    records = bmi3d.models.Decoder.objects.using(dbname).filter(**kwargs)
    return [BMI3DDecoder(r) for r in records]

'''
Filter functions
'''
def filter_has_features(features):
    '''
    Filter function to select sessions only if they had the given features enabled

    Args:
        features (list or str): a list of feature names, or a single feature

    Returns:
        function: a filter function to pass to `lookup_sessions`
    '''
    if not isinstance(features, list):
        filter_fn = lambda x: x.has_feature(features)
    else:
        filter_fn = lambda x: all([x.has_feature(f) for f in features])
    return filter_fn

def filter_has_neural_data(datasource):
    '''
    Filter function to select sessions only if they contain neural data recordings

    Args:
        datasource (str): 'ecog' or 'neuropixel'

    Returns:
        function: a filter function to pass to `lookup_sessions`

    '''
    if datasource == 'ecog':
        return lambda x: x.get_task_param('record_headstage')
    elif datasource == 'neuropixel':
        return lambda x: x.has_feature('neuropixel')
    else:
        warnings.warn('No datasource matching description')
        return lambda x: False

'''
Database objects
'''
class BMI3DTaskEntry():
    '''
    Wrapper class for bmi3d database entry classes. Written like this 
    so that other database types can implement their own classes with
    the same methods without needing to modfiy their database model.
    '''

    def __init__(self, task_entry, dbname='default'):
        '''
        Constructor

        Args:
            task_entry (TaskEntry): a database task entry from BMI3D
            dbname (str, optional): name of the database to connect to. Defaults to 'default'.
        '''
        self.dbname = dbname
        self.record = task_entry
        self.id = self.record.id
        self.datetime = self.record.date
        self.date = self.record.date.date()
        self.subject = self.record.subject.name
        self.session = self.record.session
        self.project = self.record.project

    def __repr__(self):
        '''
        Call __repr__ on the database record

        Returns:
            str: repr
        '''
        return repr(self.record)
    
    def __str__(self):
        '''
        Call __str__ on the database record

        Returns:
            str: str
        '''
        return str(self.record)
    
    @property
    def experimenter(self):
        '''
        Experimenter

        Returns:
            str: name of the experimenter
        '''
        if self.record.experimenter is None:
            return ''
        return self.record.experimenter.name
    
    @property
    def notes(self):
        '''
        Notes

        Returns:
            str: notes
        '''
        return self.record.notes

    @property
    def features(self):
        '''
        List of features that were enabled during recording

        Returns:
            list: enabled features
        '''
        return [f.name for f in self.record.feats.all()]

    @property
    def task_params(self):
        '''
        All task parameters

        Returns:
            dict: task params
        '''
        return self.record.task_params

    @property
    def sequence_params(self):
        '''
        All sequence parameters, e.g. `ntargets` or `target_radius`

        Returns:
            dict: sequence params
        '''
        return self.record.sequence_params

    @property
    def task_name(self):
        '''
        Task name, e.g. `manual control`

        Returns:
            str: task name
        '''
        return bmi3d.get_task_name(self.record, dbname=self.dbname)

    @property
    def task_desc(self):
        '''
        Task description, e.g. `flash`

        Returns:
            str: task description
        '''
        if self.record.entry_name is None:
            return ''
        return self.record.entry_name

    @property
    def duration(self):
        '''
        Duration of recording in seconds

        Returns:
            float: duration
        '''
        try:
            report = json.loads(self.record.report)
            return report['runtime']
        except:
            return 0.0
        
    @property
    def n_trials(self):
        '''
        Number of total trials presented

        Returns:
            int: number of total trials
        '''
        try:
            report = json.loads(self.record.report)
        except: 
            return 0
        total = report['n_trials']
        return total
    
    @property
    def n_rewards(self):
        '''
        Number of rewarded trials

        Returns:
            int: number of rewarded trials
        '''
        try:
            report = json.loads(self.record.report)
        except: 
            return 0
        rewards = report['n_success_trials']
        return rewards        

    def get_decoder(self, decoder_dir=None):
        '''
        Fetch the decoder object from the database, if there is one.

        Returns:
            Decoder: decoder object (type depends on which decoder is being loaded)
        '''
        if decoder_dir is not None:
            filename = bmi3d.get_decoder_name(self.record)
            filename = os.path.join(decoder_dir, filename)
        else:
            filename = bmi3d.get_decoder_name_full(self.record)
        dec = pickle.load(open(filename, 'rb'))
        dec.db_entry = self.get_decoder_record()
        return dec
    
    def get_decoder_record(self):
        '''
        Fetch the database models.Decoder record for this recording, if there is one.

        Returns:
            models.Decoder: decoder record
        '''
        return bmi3d.get_decoder_entry(self.record, dbname=self.dbname)

    def has_feature(self, featname):
        '''
        Check whether a feature was included in this recording

        Args:
            featname (str): name of the feature to check

        Returns:
            bool: whether or not the feature was enabled
        '''
        return featname in self.features

    def get_task_param(self, paramname):
        '''
        Get a specific task parameter

        Args:
            paramname (str): name of the parameter to get

        Returns:
            object: parameter value
        '''
        params = self.task_params
        if paramname not in params:
            return None
        return params[paramname]
    
    def get_sequence_param(self, paramname):
        '''
        Get a specific sequence parameter

        Args:
            paramname (str): name of the parameter to get

        Returns:
            object: parameter value
        '''
        params = self.sequence_params
        if paramname not in params:
            return None
        return params[paramname]

    def get_raw_files(self, system_subfolders=None):
        '''
        Gets the raw data files associated with this task entry 
        
        Args:
            system_subfolders (dict, optional): dictionary of system subfolders where the 
                data for that system is located. If None, defaults to the system name
        
        Returns: 
            files : list of (system, filepath) for each datafile associated with this task entry 
        '''
        return bmi3d.get_rawfiles_for_taskentry(self.record, system_subfolders=system_subfolders)

    def get_preprocessed_sources(self):
        '''
        Returns a list of datasource names that should be preprocessed for this task entry. Always
        includes experiment data (`exp`) and eye data (`eye`), and additionally includes broadband,
        lfp, and spike data if there are associated datafiles with appropriate neural data.

        Returns:
            list: preprocessed sources for this task entry
        '''
        params = self.task_params
        sources = ['exp']
        if 'record_headstage' in params and params['record_headstage']:
            sources.append('broadband')
            sources.append('lfp')
        if 'neuropixels' in self.features:
            sources.append('spike')
            sources.append('lfp')
        sources.append('eye')
        
        return sources
    
    def get_db_object(self):
        '''
        Get the raw database object representing this task entry

        Returns:
            models.TaskEntry: bmi3d task entry object
        '''
        return self.record

class BMI3DDecoder():
    '''
    Wrapper for BMI3D Decoder objects. Written like this 
    so that other database types can implement their own decoder classes with
    the same methods without needing to modfiy their database model.
    '''
    
    def __init__(self, decoder, dbname='default'):
        '''
        Constructor

        Args:
            task_entry (TaskEntry): a database task entry from BMI3D
            dbname (str, optional): name of the database to connect to. Defaults to 'default'.
        '''
        self.dbname = dbname
        self.record = decoder
        self.id = self.record.id
        if decoder.entry_id is not None:
            self.parent = lookup_sessions(id=decoder.entry_id)[0]
        else:
            self.parent = None
        self.project = self.parent.project
        self.session = self.parent.session
        self.name = self.record.name
        
    def __repr__(self):
        '''
        Call __repr__ on the database record

        Returns:
            str: repr
        '''
        if self.parent is None:
            return f"Decoder[{self.id}:{self.name}]"
        else:
            return f"Decoder[{self.id}:{self.name}] trained from {repr(self.parent)}"
    
    def __str__(self):
        '''
        Call __str__ on the database record

        Returns:
            str: str
        '''
        return str(self.record)
        
    @property
    def decoder(self):
        '''
        The decoder object

        Returns:
            object: decoder object
        '''
        if hasattr(self, 'dec'):
            return self.dec
        else:
            self.dec = self.get()
            return self.dec
        
    @property
    def channels(self):
        '''
        The decoder channels

        Returns:
            list: channels used by the decoder
        '''
        return self.decoder.channels

    @property
    def filt(self):
        '''
        The decoder filter

        Returns:
            object: decoder filter
        '''
        return self.decoder.filt

    def get(self, decoder_dir=None):
        '''
        Fetch the decoder object from the database, if there is one.

        Returns:
            Decoder: decoder object (type depends on which decoder is being loaded)
        '''
        filename = self.record.path
        if decoder_dir is not None:
            filepath = os.path.join(decoder_dir, filename)
        else:
            sys_path = bmi3d.models.System.objects.using(self.dbname).get(name='bmi').path
            filepath =  os.path.join(sys_path, filename)
        dec = pickle.load(open(filepath, 'rb'))
        self.dec = dec
        return dec

'''
Wrappers
'''
def list_entry_details(sessions):
    '''
    Returns (subject, te_id, date) for each given session.

    Args:
        sessions (list of TaskEntry): list of sessions

    Returns:
        tuple: tuple containing
            | **subject (list):** list of subject names
            | **te_id (list):** list of task entry ids
            | **date (list):** list of dates
    '''
    return zip(*[(te.subject, te.id, te.date) for te in sessions])

def group_entries(sessions, grouping_fn=lambda te: te.date):
    '''
    Automatically group together a flat list of database IDs

    Args:
        sessions (list of task entries): TaskEntry objects to group
        grouping_fn (callable, optional): grouping_fn(task_entry) takes a TaskEntry as 
            its only argument and returns a hashable and sortable object by which to group the ids

    Returns:
        list: list of tuples, each tuple containing a group of sessions
    '''
    keyed_ids = defaultdict(list)
    for te in sessions:
        key = grouping_fn(te)
        keyed_ids[key].append(te)

    keys = list(keyed_ids.keys())
    keys.sort()

    grouped_ids = []
    for date in keys:
        grouped_ids.append(tuple(keyed_ids[date]))
    return grouped_ids

def summarize_entries(entries, sum_trials=False):
    '''
    Generates a dataframe summarizing the subject, date, task, number of trials, 
    and duration in minutes of each entry in the input list. Optionally sum the
    number of trials and duration for unique tasks across days for each subject
    
    Args:
        entries (list): list of bmi3d task entries
        sum_trials (bool, optional): sum the number of trials and duration across
            unique tasks for each day for each subject

    Returns:
        pd.DataFrame: dataframe of entry summaries

    Examples:

        .. code-block:: python

            date_obj = date.fromisoformat('2023-02-06')
            entries = db.lookup_sessions(date=date_obj)
            df = db.summarize_entries(entries)
            display(df)

        .. image:: _images/db_summarize_sessions.png

        .. code-block:: python

            df_unique = db.summarize_entries(entries, sum_trials=True)
            display(df_unique)

        .. image:: _images/db_summarize_sessions_sum.png

    '''

    # Generate a summary dataframe
    desc = {
        'subject': [e.subject for e in entries],
        'te_id': [e.id for e in entries],
        'date': [e.date for e in entries],
        'time': [e.datetime.time().replace(microsecond=0) for e in entries],
        'task_name': [e.task_name for e in entries],
        'task_desc': [e.task_desc for e in entries],
        'n_rewards': [e.n_rewards for e in entries],
        'n_trials': [e.n_trials for e in entries],
        'duration_minutes': [np.round(e.duration/60, 1) for e in entries],
    }
    all_sessions = pd.DataFrame(desc)
    if sum_trials is False:
        return all_sessions

    # Optionally sum the dataframe across unique tasks on each day for each subject
    unique_sessions = all_sessions.drop('te_id', axis=1).groupby(
        ['subject', 'date', 'task_name', 'task_desc']).sum(numeric_only=True)
    return unique_sessions



import os
import numpy as np

from .base import get_preprocessed_filename, load_preproc_eye_data, save_hdf, find_preproc_ids_from_day
from ..postproc import get_calibrated_eye_data

def proc_eye_day(preproc_dir, subject, date, correlation_min=0.9, dry_run=False):
    '''
    Finds files from the given subject and date with the best eye calibration and automatically 
    applies it to every recording on that day for that subject. If no good calibration is found,
    raises a ValueError exception.
    
    Args:
        preproc_dir (str): base directory where the files live
        subject (str): Subject name
        date (str): Date of recording
        correlation_min (float, optional): correlation below which is unacceptable
        dry_run (bool, optional): if True, files will not be modified. 
        
    Raises:
        ValueError

    Returns:
        tuple: tuple containing:
            | **best_id (int)**: the task entry id with the highest mean absolute value correlation coefficient
            | **te_ids (list of int)**: the ids to which the coeff were applied
    '''
    
    # Find best calibration from the given subject and date 
    te_ids = find_preproc_ids_from_day(preproc_dir, subject, date, 'eye')
    if len(te_ids) == 0:
        print(f"No preprocessed files found on {date}")
        return None, []
    best_id = None
    best_coeff = None
    best_correlation = 0
    for te_id in te_ids:
        eye_data, eye_metadata = load_preproc_eye_data(preproc_dir, subject, te_id, date)
        if 'external_calibration' in eye_metadata and eye_metadata['external_calibration']:
            continue # ignore the file if it has already had another calibration applied to it
        if 'correlation_coeff' not in eye_data:
            continue # ignore if there isn't any calibration data
        correlation = np.mean(abs(eye_data['correlation_coeff']))
        print(correlation)
        if correlation > best_correlation and correlation < 1.0: # 1.0 correlations aren't valid
            best_id = te_id
            best_coeff = eye_data['coefficients']
            best_correlation = correlation
            best_version = eye_metadata.get('calibration_version', 'unknown')
            best_date = eye_metadata.get('calibration_date', '2000-01-01')
    
    if best_correlation < correlation_min:
        raise ValueError(f"Could not find calibrated eye data with correlation > {correlation_min}"
                         f" for {subject} on {date} (best {best_correlation})")
        
    # Apply that calibration to all the other files
    te_ids = np.delete(te_ids, np.where(te_ids == best_id)[0])
    print(f"Applying eye calibration from {subject} block {best_id} on {date} (r={best_correlation})"
          f" to {len(te_ids)-1} files...")
    if dry_run:
        return best_id, te_ids

    for te_id in te_ids:
        eye_data, eye_metadata = load_preproc_eye_data(preproc_dir, subject, te_id, date)
        eye_data['calibrated_data'] = get_calibrated_eye_data(eye_data['raw_data'], best_coeff)
        eye_data['coefficients'] = best_coeff
        eye_metadata['external_calibration'] = True
        eye_metadata['external_calibration_id'] = best_id
        eye_metadata['calibration_version'] = best_version
        eye_metadata['calibration_date'] = best_date
        preproc_file = get_preprocessed_filename(subject, te_id, date, 'eye')
        preproc_dir_subject = os.path.join(preproc_dir, subject)
        save_hdf(preproc_dir_subject, preproc_file, eye_data, "/eye_data", append=True)
        save_hdf(preproc_dir_subject, preproc_file, eye_metadata, "/eye_metadata", append=True)

    return best_id, te_ids



# tutorial_functions.py
# This is an example module to practice using git.

# Ryan's practice function
def practice_func_ryan(user_name):
    '''
    This function returns a print statement specific to Ryan from the input username

    Inputs:
        user_name [string]: Name of person running the function
        
    Outputs:
        string_out [string]: Hello Ryan from 
    '''
    string_out = 'Hello Ryan, from' + user_name
    return string_out

# Ryan's practice function 2
def practice_func_ryan2(favorite_food):
    '''
    This function returns a print statement specific to Ryan from the input username

    Inputs:
        favorite_food [string]: Your favorite food
        
    Outputs:
        None; print statement
    '''
    string_out = 'My favorite food is' + favorite_food
    return string_out

# Leo's practice function
def practice_func_leo(number_of_fish):
    '''
    This function prints a number of fish

    Inputs:
        number_of_fish [int]: How many fish you have
        
    Outputs:
        statement [str]: a statement of how many fish you have
    '''
    
    return 'You have {} fish'.format(number_of_fish)


#Pamel's practice function
def pamels_function(num_of_cookies):
    ''' This function prints the number of cookies Pamel has

    Inputs: num_of_cookies [int]: How many cookies Pamel has

    Outputs:
        statement[str]: a statement about my cookies
    '''
    return 'Pamel has {} cookies. She will eat all of them'.format(num_of_cookies)

def practice_function_gus(favorite_food):
    '''
    This function returns a print statement specific to Gus from the input username

    Inputs:
        favorite_food [string]: your favorite food

    Outputs:
        string_out [string]: returns a string saying what your favorite food is
    '''
    string_out = 'my favorite food is ' + favorite_food
    return string_out

  
def practice_func_tomo(user_name):
    '''
    This function returns a print statement specific to Tomo from the input username

    Inputs:
        user_name [string]: Name of person running the function
        
    Outputs:
        string_out [string]: Hello Tomo from 
    '''
    string_out = 'Hello Tomo, from ' + user_name
    return string_out


def practice_func_pavi(my_name):
    '''
    This is just a test function
    Inputs:
      param my_name: [str] enter your name
    
    Output:
      return: returns a statement introducing my_name
    '''

    return 'hello! I am '+ my_name


def practice_func_sijia(favourite_marvel_character):
    '''
    This function prints a number of fish

    Inputs:
        favourite_marvel_character [string ]: which marvel character we'd like to meet
        
    Outputs:
        statement [str]: a greeting statement to the favourite marvel character
    '''
    
    return f'hello, {favourite_marvel_character}'


def display_cat_leo():
    '''
    Cat

    Inputs:
        None

    Outputs:
        cat [str]: a cat
    '''
    cat = ' /\\_/\\\n( o.o )\n > ^ <'
    for t in range(3):
        print(cat)
    return cat

def practice_func_miken(your_number):
    '''
    This function tests if your number is the best number

    Inputs:
        your_number [num]: your number, any number.

    Outputs:
        result [bool]: True if your number is the best number, otherwise false.
    '''

    result = your_number == 7
    return result 


def practice_function_gus(favorite_food):
    '''
    This function returns a print statement specific to Gus from the input username

    Inputs:
        favorite_food [string]: your favorite food

    Outputs:
        string_out [string]: returns a string saying what your favorite food is
    '''
    string_out = 'my favorite food is ' + favorite_food
    return string_out


# torch.py
# code using the pytorch library which isn't installed by default. to install use `pip install torch`

import torch

# - - -- --- ----- -------- ---PYTORCH--- -------- ----- --- -- - -
# - - -- --- ----- -------- ---DATASETS-- -------- ----- --- -- - -

class TensorDataset(torch.utils.data.Dataset):
    r"""Dataset wrapping tensors. Implements the pytorch Dataset parent class.

    Each sample will be retrieved by indexing tensors along the first dimension. These samples are collected and returned in a list of tensors in the same order as their source tensors.

    Args:
        *tensors (Tensor): pytorch tensors that have the same size of the first dimension. This first dimension indexes over individual trials. Standard convention has the 2nd dimension index over time samples while extra dimensions index over channels or spatial dimentions.
    
    Parameters:
        device (str): Memory location to place input tensors. (default: 'cpu')
        transform (torch.nn.Module): Transform class for augmenting or transforming tensor samples. (default: None)
        transform_mast ([bool]): List of boolean values indicating which input tensors are transformed when sampled. (default: None)

    Returns:
        sample ([Tensor]): List of pytorch Tensors sampled from an input index (int).
    """

    def __init__(self, *tensors, device='cpu', transform=None, transform_mask=None):
        assert all(tensors[0].shape[0] == tensor.shape[0] for tensor in tensors)
        self.tensors = tensors
        self.device = device
        self.transform = transform
        if transform_mask:
            assert len(self.tensors) == len(transform_mask), f'transform_mask length ({len(transform_mask)}) must match number of tensors ({len(tensors)}).'
        else:
            transform_mask = [True] * len(self.tensors) # all-hot mask
        self.transform_mask = transform_mask


    def __getitem__(self, index):
        # get samples
        sample = [tensor[index] for tensor in self.tensors]
        # apply transform
        if self.transform:
            for idx, s in enumerate(sample):
                if self.transform_mask[idx]:
                    sample[idx] = self.transform(s)
        # assign device
        sample = recursive_assign_device(sample,self.device)
        return sample

    def __len__(self):
        return self.tensors[0].size(0)

def recursive_assign_device(x, device: str):
    """Recursively assign tensor elements in a nested list or tuple of arbitrary depth to a specified device memory location.

    Args:
        x ([Tensor,[Tensor],...]): List of tensors or lists of tensors. Can be arbitrarily deep.
        device (str): Memory location on the current system (ex: 'cpu', 'cuda:0')

    Returns:
        [Tensor,[Tensor],...]: Copy of input x, memory relocated to designated location.
    """
    if isinstance(x,(list,tuple)):
        x = [recursive_assign_device(_x,device) for _x in x]
    else:
        x = x.to(device)
    return x



import numpy as np
import types
from .arraymethods import *

class ChunkedStream:
    def __init__(self, chunkemitter=None):
        if chunkemitter is None:
            raise StopIteration('No dataset emitter source specified.')
        
        if not isinstance(chunkemitter, types.GeneratorType):
            raise TypeError('ChunkedStream must be initialized on chunk emitters.')

        self.chunkemitter = iter(chunkemitter)

    def __next__(self):
        return next(self.chunkemitter)

    def __iter__(self):
        return self

    def find_all_nonzero(self, chmask=None, discontchunklength=0.1, threshold=None, multiplier=None):

        for emitchunk in self:
            if chmask is not None:
                chunk = emitchunk[chmask, :]
            else:
                chunk = emitchunk

            disconts = np.nonzero(chunk != 0)

            print("{} discontinuities found in block".format(len(disconts[1])))
            yield(disconts[0], disconts[1], chunk[disconts])

    def find_unique_discont(self, chmask=None, discontchunklength=0.1, threshold=None, multiplier=None):
        discontlist = []
        discontchunklist = []
        windowsize = int(discontchunklength*25000/2)

        for emitchunk in self:
            if chmask is not None:
                chunk = emitchunk[chmask, :]
            else:
                chunk = emitchunk

            if chunk.metadata['isstart'] is True:
                # start of a sequence
                t_chunk_offset = 0
                disconts, state = find_discont(chunk, threshold=threshold, multiplier=multiplier)
                windowsize = int(discontchunklength*chunk.metadata['samplerate']/2)
            else:
                disconts, state = find_discont(chunk, threshold=threshold, multiplier=multiplier, contstate=state)

            chunkdisconts = unique_disconts(disconts)
            print("Discontinuities: {}".format(chunkdisconts) if len(chunkdisconts) > 0 else "... no discontinuities found.")
            if len(chunkdisconts) > 0:
                for pt in chunkdisconts:
                    # index within chunk (indexable to data)
                    pt_ch = pt[0]
                    pt_t = pt[1]

                    # index within whole sequence, across multiple chunks
                    # (cannot be used, for printing information only)
                    pt_print = (pt_ch, pt_t + t_chunk_offset)

                    t_from = max(0, pt_t - windowsize)
                    t_to = min(chunk.shape[1]-1, pt_t + windowsize)
                    ptchunk = chunk[pt_ch, t_from:t_to]
                    yield pt_print, ptchunk

            discontlist.extend(chunkdisconts)
            t_chunk_offset += chunk.shape[1]

        return discontlist

def find_discont(indata, threshold=None, multiplier=None, contstate=None):
    if threshold is None:
        threshold = 99
    if multiplier is None:
        multiplier = 2

    if contstate is not None:
        prevrow = contstate["data"]
        if prevrow.shape[0] == indata.shape[0] and prevrow.shape[1] == 1:
            data = np.hstack((prevrow, indata))
        else:
            raise TypeError('contstate["data"] must be [channels x 1]')

        prevthres = contstate["thresval"]
    else:
        data = indata.astype(int)
        prevthres = np.inf

    diffs = np.diff(data)
    thresval = np.percentile(abs(diffs), threshold)*multiplier
    thresval = min(thresval, prevthres)

    discont_tuples = np.nonzero(abs(diffs) > thresval)
    timesortidx = np.argsort(discont_tuples[1])

    retstate = {"data": data[:,-1:], "thresval": thresval}

    return (discont_tuples[0][timesortidx], discont_tuples[1][timesortidx], diffs[discont_tuples]), retstate

def unique_disconts(discont_tuples, dup_dt=500):
    uniq_discont = []
    discont_ch = discont_tuples[0]
    discont_t = discont_tuples[1]

    if len(discont_ch) != len(discont_t):
        raise IndexError('Number of channels and timepoints do not match')

    if len(discont_t) > 0:
        # find a list of time changes greater than dup_dt samples (250 samples==1ms)
        # collapse ch listing if <= dup_dt samples
        discont_changepts = np.nonzero(np.diff(discont_t) > dup_dt)[0]
        # 0 gets first element of of discont_t
        # discont_changepts + 1 gets element after change occurs in discont_t
        discont_t_list = np.hstack((0,discont_changepts+1))

        # iterate through discont_t_list
        for t_list_i in range(len(discont_t_list)):
            # t is the smallest continuous block of time
            t = discont_t[discont_t_list[t_list_i]]

            # if there are more discont_t_list elements after this one, current t captures
            # all ch between this one and the next one
            if t_list_i < len(discont_t_list)-1:
                ch = np.unique(discont_ch[discont_t_list[t_list_i]:discont_t_list[t_list_i+1]])
            # if this is the last discont_t_list element (or the only one), capture
            # all the rest of the ch
            else:
                ch = np.unique(discont_ch[discont_t_list[t_list_i]:])

            # add to the list of discontinuities
            uniq_discont.append((ch, t))
    return uniq_discont